from magbox import llg,spin, Lattice, Vars
from torch import sin, cos
from matplotlib import pyplot as plt
import numpy as np
import cProfile, pstats, io
# from pyinstrument import Profiler

pr=cProfile.Profile()
pr.enable()

# trace(files=files.all)

# pr=Profiler()
# pr.start()

rng=np.random.default_rng()

N1=128
N2=128
lt = Lattice(type="square", size=[N1, N2], periodic=True)
vars = Vars(J=0)
theta0=np.acos(rng.random([N1,N2])*2-1)
# theta0=np.ones([N1])*0.1
phi0=rng.random([N1,N2])*2*np.pi
# theta0=np.ones([N1])*0.1
# phi0=np.zeros([N1])
alpha=0.1


sp=spin(theta0,phi0,lattice_type=lt,device="cpu",type='f64')
sf=llg(sp,vars=vars, dt=0.1,alpha=alpha,T=1,Temp=0.1)
ang=sp.ang

t,ang,stats,erro_info=sf.run(sp)

en=sf.h_fun.energy(ang)

t_np=t.cpu().detach().numpy()
en_np=en.cpu().detach().numpy()
en_np=en_np/N1

pr.disable()
s = io.StringIO()
sortby = "cumtime"  
ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
ps.print_stats()
# print(s.getvalue())
pr.dump_stats("llg_thermal_cpu_f64.prof")

# pr.stop()
# pr.open_in_browser()

plt.figure(figsize=(6,6))

plt.subplot(2,1,1)
plt.plot(t_np,en_np,label="energy")
plt.xlabel("Time")
plt.title("energy")
plt.legend()

plt.tight_layout()
plt.show()





import torch

crow_indices = torch.tensor([0, 2, 4])
col_indices = torch.tensor([0, 1, 0, 2])
values1 = torch.tensor([[[0, 1, 2], [6, 7, 8]],
                       [[3, 4, 5], [9, 10, 11]],
                       [[12, 13, 14], [18, 19, 20]],
                       [[15, 16, 17], [21, 22, 23]]])
values2 = torch.tensor([[[0, 1,], [6, 7]],
                       [[3, 4], [9, 10]],
                       [[12, 13], [18, 19]],
                       [[15, 16], [21, 22]]])
bsr1 = torch.sparse_bsr_tensor(crow_indices, col_indices, values1, dtype=torch.float64,device=torch.device("cuda"))
bsr2= torch.sparse_bsr_tensor(crow_indices, col_indices, values2, dtype=torch.float64,device=torch.device("cpu"))
v1=torch.tensor([[1],[2],[3],[4],[5],[6],[4],[5],[6]],dtype=torch.float64,device=torch.device("cuda"))
v2=torch.tensor([[1],[2],[3],[4],[5],[6]],dtype=torch.float64,device=torch.device("cpu"))
# print(bsr1 @ v1)
print(bsr2 @ v2)


import torch
from typing import Callable, Tuple, Dict, Any, Optional
from tqdm import tqdm
import time
import warnings
import math
from .Wait_bar import Wait_bar
from .initial import Lattice, Vars
from .spin import spin

def get_data_type(type):
    if type=="f32":
        data_type=torch.float32
    elif type=="f64":
        data_type=torch.float64
    elif type=='f16':
        data_type=torch.float16
    else:
        raise ValueError("type must be f16, f32 or f64")
    return data_type
    
def get_device(device):
    if device=="gpu":
        if torch.cuda.is_available():
            return torch.device("cuda")
        else:
            print("CUDA is not available, using CPU instead.")
            return torch.device("cpu")
    elif device=="cpu":
        return torch.device("cpu")
    else:
        raise ValueError("device must be 'cpu' or 'gpu'")
def get_Jmtx(lattice_type: Lattice,device=torch.device("cuda"),data_type=torch.float32) -> torch.Tensor:
    l_type=lattice_type.type

    if l_type=="square":
        N=lattice_type.size
        N_dim=len(N)
        if N_dim==1:
            N=N+[1,1]
        elif N_dim==2:
            N=N+[1]
            
        pd0=lattice_type.periodic
        if is_bool_or_single_bool_list(pd0):
            pd=create_bool_list(pd0, N)
            if N_dim==1:
                pd[1]=False
                pd[2]=False
            elif N_dim==2:
                pd[2]=False
        direction=lattice_type.J_direction
        totalN=math.prod(N)
        N1=N[0]
        N2=N[1]
        N3=N[2]
        if direction is None:
            # 所有方向的耦合
            v = torch.ones(3 * totalN - N1 * N2 - N2 * N3 - N3 * N1, dtype=data_type,device=device) / 2
            
            # get backward coupling, direction 1
            i = torch.arange(1, totalN)  
            back_boundary = (i % N1 == 0)
            i = i[~back_boundary]
            j = i.clone()  
            
            # get right coupling, direction 2
            itmp = torch.arange(1, totalN)
            right_boundary = ((itmp - 1) % (N1 * N2) - N1 * (N2 - 1)) >= 0
            itmp = itmp[~right_boundary]
            i = torch.cat([i, itmp])
            j = torch.cat([j, itmp + N1 - 1])  # 调整索引
            
            # get bottom coupling, direction 3
            i = torch.cat([i, torch.arange(1, totalN - N1 * N2 + 1)])
            j = torch.cat([j, torch.arange(N1 * N2 , totalN )])
            
            if pd[0]:  # periodic boundary condition in direction 1
                back_forward_i = torch.arange(1, totalN + 1, N1)
                back_forward_j = torch.arange(N1-1, totalN, N1)
                i=torch.cat([i, back_forward_i])
                j=torch.cat([j, back_forward_j])
                v= torch.cat([v, torch.ones(N2 * N3, dtype=data_type,device=device) / 2])
            if pd[1]:  # periodic boundary condition in direction 2
                left_right_i = torch.arange(1, totalN + 1)
                tmpbd = ((left_right_i - 1) % (N1 * N2)) >= N1
                left_right_i = left_right_i[~tmpbd]
                left_right_j = torch.arange(0, totalN)
                tmpbd = (left_right_j  % (N1 * N2)) < N1 * (N2 - 1)
                left_right_j = left_right_j[~tmpbd]
                i=torch.cat([i, left_right_i])
                j=torch.cat([j, left_right_j])
                v= torch.cat([v, torch.ones(N1 * N3, dtype=data_type,device=device) / 2])
            if pd[2]:  # periodic boundary condition in direction 3
                up_down_i = torch.arange(1, N1 * N2 + 1)
                up_down_j = torch.arange(N1 * N2 * (N3 - 1), totalN)
                i=torch.cat([i, up_down_i])
                j=torch.cat([j, up_down_j])
                v= torch.cat([v, torch.ones(N1 * N2, dtype=data_type,device=device) / 2])

        else:
            # 只有一个方向的耦合
            if direction==0: # backward耦合（x方向）
                i = torch.arange(1, totalN)
                back_boundary = (i % N1 == 0)
                i = i[~back_boundary]
                j = i.clone()  # 调整索引
                v = torch.ones(len(i), dtype=data_type,device=device) / 2
                if pd[0]:  # periodic boundary condition in direction 1
                    back_forward_i = torch.arange(1, totalN + 1, N1)
                    back_forward_j = torch.arange(N1-1, totalN , N1)
                    i=torch.cat([i, back_forward_i])
                    j=torch.cat([j, back_forward_j])

                v = torch.ones(len(i), dtype=data_type,device=device) / 2
            elif direction==1: # right耦合（y方向）
                i = torch.arange(1, totalN)
                right_boundary = ((i - 1) % (N1 * N2) - N1 * (N2 - 1)) >= 0
                i = i[~right_boundary]
                j = i + N1 - 1  # 调整索引
                if pd[1]:  # periodic boundary condition in direction 2
                    left_right_i = torch.arange(1, totalN + 1)
                    tmpbd = ((left_right_i - 1) % (N1 * N2)) >= N1
                    left_right_i = left_right_i[~tmpbd]
                    left_right_j = torch.arange(0, totalN )
                    tmpbd = (left_right_j % (N1 * N2)) < N1 * (N2 - 1)
                    left_right_j = left_right_j[~tmpbd]
                    i=torch.cat([i, left_right_i])
                    j=torch.cat([j, left_right_j])
                v = torch.ones(len(i), dtype=data_type,device=device) / 2
            elif direction == 2:  # bottom耦合（z方向）
                i = torch.arange(1, totalN - N1 * N2 + 1)
                j = torch.arange(N1 * N2, totalN)
                if pd:  # 周期性边界条件
                    up_down_i = torch.arange(1, N1 * N2 + 1)
                    up_down_j = torch.arange(N1 * N2 * (N3 - 1), totalN)
                    
                    i = torch.cat([i, up_down_i])
                    j = torch.cat([j, up_down_j])
                v = torch.ones(len(i), dtype=data_type,device=device) / 2
            else:
                raise ValueError('direction must be 1, 2, 3 or None')
        i = i - 1  # Convert to 0-based index
        Jmtx=torch.sparse_coo_tensor(torch.stack([i, j]), v, (totalN, totalN),dtype=data_type,device=device)
       
    return Jmtx+Jmtx.t()
def is_bool_or_single_bool_list(x):
    if isinstance(x, bool):
        return True
    elif isinstance(x, list) and len(x) == 1 and isinstance(x[0], bool):
        return True
    return False
def create_bool_list(x, y) -> list[bool]:
    """创建与y同长的布尔列表"""
    # 获取实际的布尔值
    if isinstance(x, bool):
        bool_val = x
    elif isinstance(x, list) and len(x) == 1 and isinstance(x[0], bool):
        bool_val = x[0]
    else:
        raise ValueError("x必须是布尔值或单元素布尔值列表")
    
    return [bool_val] * len(y)
class eq_solver:
    def __init__(self,odeFcn, t_span, y0, solver_name, options):
        self.device=y0.device
        self.dtype=y0.dtype
        self._ode_options(options)
        self._ode_initial(odeFcn, t_span, y0)
        self._tableau(solver_name)
    def _ode_options(self, options):
        device=self.device
        dtype=self.dtype
        if options is None:
            options = {}
        
        # Initialize options
        self.waitbar = options.get('waitbar', True)
        
        # Extract odeset options
        rtol = options.get('rel_tol', torch.tensor(1e-3,device=device,dtype=dtype))
        atol = options.get('abs_tol', torch.tensor(1e-6,device=device,dtype=dtype))
        max_failures = options.get('max_consecutive_failures', torch.tensor(10,device=device,dtype=torch.int64))
        refine = options.get('refine', 4)
        max_step = options.get('max_step', torch.tensor(2**32-1,device=device, dtype=torch.int64))

        self.rtol=torch.as_tensor(rtol, device=device, dtype=dtype)
        self.atol=torch.as_tensor(atol, device=device, dtype=dtype)
        self.max_failures=torch.as_tensor(max_failures, device=device, dtype=torch.int64)
        self.refine=refine
        self.max_step = torch.as_tensor(max_step, device=device, dtype=torch.int64)

    def _ode_initial(self, odeFcn, t_span, y0):
        device=self.device
        dtype=self.dtype
        self.t0 = t_span[0]
        self.t_final = t_span[-1]
        self.t_dir = torch.sign(self.t_final - self.t0)
        # step size constraints
        h_min = 16 * torch.finfo(dtype).eps
        h_min=torch.tensor(h_min, dtype=dtype,device=device)
        safe_h_max = 16.0 * torch.finfo(dtype).eps * torch.max(torch.abs(self.t0), torch.abs(self.t_final))
        default_h_max = torch.max(0.1 * torch.abs(self.t_final - self.t0), safe_h_max)
        h_max = torch.min(torch.abs(self.t_final - self.t0), self.max_step)

        t = self.t0.clone()
        y = y0.clone()
        y=y.view(-1,1)


        n_t_span=t_span.shape[0]
        n_eq=y0.shape[0]
        S=torch.tensor(0,device=device,dtype=dtype)
        chunk=0
        refine=self.refine
        if n_t_span > 2:
            output_pos = 1  # output at t_span points
        elif self.refine <= 1:
            output_pos = 2  # computed points
        else:
            output_pos = 3  # computed points, with refinement
            S = torch.linspace(1/refine, 1 - 1/refine, refine - 1, dtype=dtype, device=device)
        # Initialize output arrays
        if n_t_span > 2:
            t_out = torch.zeros(n_t_span, dtype=dtype, device=device)
            y_out = torch.zeros(n_eq, n_t_span, dtype=dtype, device=device)
        else:
            chunk = min(max(100, 50 * refine), refine + (2**13) // n_eq)
            t_out = torch.zeros(chunk, dtype=dtype, device=device)
            y_out = torch.zeros(n_eq, int(chunk), dtype=dtype, device=device)

        self.S=S
        self.chunk=chunk
        self.h_min=h_min
        self.h_max=h_max
        self.t=t
        self.y=y
        self.n_t_span=n_t_span
        self.n_eq=n_eq
        self.t_span=t_span
        self.output_pos=output_pos
        self.t_out=t_out
        self.y_out=y_out
        self.ode_fcn=odeFcn

        # return t0, t_final, t_dir, y0, h_min, h_max, t, y0, output_pos, t_out, y_out, chunk, S
    def _tableau(self, solver_name):
        device=self.device
        dtype=self.dtype
        if solver_name == 'RK45':
            self.alpha=torch.tensor([1 / 5, 3 / 10, 4 / 5, 8 / 9, 1., 1.], 
                                    dtype=dtype,device=device)
            self.beta=[
                torch.tensor([1 / 5], dtype=dtype,device=device).view(-1,1),
                torch.tensor([3 / 40, 9 / 40], dtype=dtype,device=device).view(-1,1),
                torch.tensor([44 / 45, -56 / 15, 32 / 9], dtype=dtype,device=device).view(-1,1),
                torch.tensor([19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729], dtype=dtype,device=device).view(-1,1),
                torch.tensor([9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656], dtype=dtype,device=device).view(-1,1),
                torch.tensor([35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84], dtype=dtype,device=device).view(-1,1)
            ]
            self.c_error=torch.tensor([71/57600, 0.0, -71/16695, 71/1920, -17253/339200, 22/525, -1/40]
                # [71 / 86400, 0, -142 / 50085, 71 / 2880, -5751 / 169600, 44 / 1575, -1 / 60],
                                      , dtype=dtype,device=device).view(-1,1)
            self.order=torch.tensor(5, dtype=dtype,device=device)
            self.interp_coeff=torch.tensor([
                [1, -183/64, 37/12, -145/128],
                [0,0,0,0],
                [0,1500/371, -1000/159, 1000/371],
                [0, -125/32, 125/12, -375/64],
                [0, 9477/3392, -729/106, 25515/6784],
                [0,-11/7, 11/3, -55/28],
                [0, 3/2, -4, 5/2]
            ], dtype=dtype,device=device)
        else:
            raise ValueError(f'Unknown solver: {solver_name}')
    def run(self,bar: Wait_bar):
        finished = False

        next_idx = 1 # for t_span output
        n_failures = 0
        integration_failed= False
        err_history=[]
        n_steps = 0

        n_out=0

        t_out=self.t_out
        y_out=self.y_out
        t=self.t
        y=self.y
        dtype=self.dtype
        device=self.device
        h_max=self.h_max
        h_min=self.h_min
        rtol=self.rtol
        atol=self.atol
        max_failures=self.max_failures
        t_final=self.t_final
        t0=self.t0
        t_dir=self.t_dir
        t_span=self.t_span
        waitbar=self.waitbar
        alpha=self.alpha
        beta=self.beta
        c_error=self.c_error
        ode_fcn=self.ode_fcn
        S=self.S
        chunk=self.chunk
        output_pos=self.output_pos
        n_t_span=self.n_t_span
        n_eq=self.n_eq
        refine=self.refine
        order=self.order
        interp_coeff=self.interp_coeff

        f1=ode_fcn(t,y).view(-1,1)
        
        n_calls=1

        t_out[n_out]=t
        y_out[:,n_out]=y.view(-1)

        t2pi=torch.tensor(2*math.pi,dtype=dtype,device=device)
        h = torch.min(h_max, torch.max(h_min, 0.1 * torch.abs(t_final -t0)))
        h_abs=torch.abs(h)

        while not finished:
            h_abs=torch.min(h_max, torch.max(h_min, h_abs))
            h = t_dir*h_abs
            if h_abs > torch.abs(t_final-t):
                h = t_final - t
                h_abs = torch.abs(h)
                finished = True
            failed = False
            y_list=y.clone()
            f_list=f1.clone()
            while True:
                t_new = t + h
                for i, (alpha_i, beta_i) in enumerate(zip(alpha,beta)):
                    y_list=torch.cat([y_list, y + h * f_list @ beta_i],dim=1)
                    if alpha_i == 1. :
                        ti=t_new
                    else:
                        ti=t+alpha_i*h
                    f_list=torch.cat([f_list,ode_fcn(ti,y_list[:,-1:])],dim=1)
                n_calls += 6
                y_new = y_list[:,-1:]
                err = f_list @ c_error

                err /= torch.max(torch.max(y.abs(),y_new.abs()), atol)
                err = h_abs *  torch.max(err.abs())
                err=err.item()
                # step acceptance
                accept_step=err <= rtol
                if accept_step:
                    n_failures = 0
                if h <= h_min:
                    accept_step = True
                    n_failures +=1
                    failed=True
                    if n_failures >= max_failures:
                        bar.close(waitbar)
                        warnings.warn(
                            f"Step size reached minimum hmin = {h_min.item():.2e} at t={t.item():.2e}, but still cannot satisfy tolerance. "
                            f"Current error: {err:.2e}, Required tolerance: {rtol:.2e}. "
                            f"This may indicate a stiff ODE or overly strict tolerances. "
                            f"Consider using a stiff solver or relaxing tolerances.",
                            RuntimeWarning,
                            stacklevel=2
                        )
                        finished = True
                        integration_failed = True
                        break
                else:
                    n_failures = 0 # Reset if we're still above hmin
                if accept_step:
                    err_history.append(err)
                    break
                else:
                    if failed:
                        h_abs = torch.max(h_min, 0.5* h_abs)
                    else:
                        failed = True 
                        h_abs =step_after_nofailed(h_min,h_abs,rtol,err,order)
                    h = t_dir * h_abs
                    y_list=y.clone()
                    f_list=f1.clone()
                    finished = False
            n_steps += 1
            if integration_failed:
                break
            # Update waitbar if enabled
            bar.update(t_new, waitbar, finished)
            # output
            if output_pos ==2: # computed points
                nout_new = 1
                t_out_new = t_new.unsqueeze(0)
                y_out_new = y_new.view(-1,1)
            elif output_pos ==3: # computed points, with refinement
                t_ref = t + (t_new - t) * S
                nout_new = refine
                t_out_new = torch.cat([t_ref, t_new.unsqueeze(0)])
                y_interp = interp_fun(t_ref, t, y, h, f_list,interp_coeff)
                y_out_new = torch.cat([y_interp, y_new.view(-1,1)], dim=1) 
            else:
                nout_new = 0
                t_out_new = torch.tensor([], dtype=dtype, device=device)
                y_out_new = torch.tensor([], dtype=dtype, device=device)
                
                while next_idx < n_t_span:
                    if t_dir * (t_new - t_span[next_idx]) < 0:
                        break
                    nout_new += 1
                    t_out_new = torch.cat([t_out_new, t_span[next_idx].unsqueeze(0)])
                    if t_span[next_idx] == t_new:
                        y_out_new = torch.cat([y_out_new, y_new], dim=1)
                    else:
                        y_interp = interp_fun(t_span[next_idx], t, y, h, f_list, interp_coeff)
                        y_out_new = torch.cat([y_out_new, y_interp], dim=1)
                    next_idx += 1
            # y_out_new = y_out_new % t2pi
            # Store output
            if nout_new > 0:
                old_nout = n_out
                n_out += nout_new

                if n_out+1 > t_out.shape[0]:
                    extra = max(chunk, nout_new)
                    tout_new_temp = torch.zeros(t_out.shape[0] + extra, dtype=dtype, device=device)
                    tout_new_temp[:t_out.shape[0]] = t_out
                    t_out = tout_new_temp
                    
                    yout_new_temp = torch.zeros(n_eq, y_out.shape[1] + extra, dtype=dtype, device=device)
                    yout_new_temp[:, :y_out.shape[1]] = y_out
                    y_out = yout_new_temp
            
                t_out[old_nout+1:n_out+1] = t_out_new
                y_out[:, old_nout+1:n_out+1] = y_out_new
            h_abs = _optimal_step_size(h_abs, err/rtol, order, failed)
        
            t=t_new
            y=y_new
            # y=y % t2pi
            f1 = f_list[:,-1:]

        bar.close(waitbar)

        t_out= t_out[:n_out+1]
        y_out = y_out[:,:n_out+1]
        stats = {'n_calls': n_calls,
            'n_steps': n_steps,
            'n_output': n_out+1,
            'intergration': not integration_failed}
        err_info = {
            'err_history': err_history,
            'max_step_error': max(err_history) if err_history else 0.0
        }
        return t_out, y_out, stats, err_info
def interp_fun(t_interp: torch.Tensor, t: torch.Tensor, y: torch.Tensor,h: torch.Tensor ,f_list: torch.Tensor, interp_coeff)-> torch.Tensor:
    """
    Interpolation function for Dormand-Prince method.
    """
    max_order = interp_coeff.shape[1] # interp_coeff.shape = [f_order, t_order]
    s = (t_interp - t) / h
    s = s.reshape(1,-1) #[1,t_l]
    s_list = s 
    if max_order > 1:
        for jj in range(max_order-1):
            s_list=torch.cat([s_list,s**(jj+2)],dim=0) # [t_order,t_l]
    s_coeff = interp_coeff @ s_list #[f_order, t_order] @ [t_order,t_l] = [f_order, t_l]
    y_interp = y + h* f_list@ s_coeff  # [y_d, f_order] @ [f_order, t_l] =[y_d,t_l]
    return y_interp

def step_after_nofailed(h_min,h_abs,rtol,err,order):
    return torch.max(h_min, h_abs * max(0.1, 0.8 * (rtol / err) ** (1/order)))

def _optimal_step_size(last_step, error_ratio, order, failedflag, safety=0.8, ifactor=4.0, dfactor=0.2):
    """Calculate the optimal size for the next step."""
    if failedflag:
        return last_step
    else: 
        ifactor=torch.as_tensor(ifactor,dtype=last_step.dtype,device=last_step.device)
        dfactor=torch.as_tensor(dfactor,dtype=last_step.dtype,device=last_step.device)
        if error_ratio == 0:
            return last_step * ifactor
        if error_ratio < 1:
            dfactor = torch.ones((), dtype=last_step.dtype, device=last_step.device)
        error_ratio = error_ratio.type_as(last_step)
        exponent = torch.as_tensor(order, dtype=last_step.dtype, device=last_step.device).reciprocal()
        factor = torch.min(ifactor, torch.max(safety / error_ratio ** exponent, dfactor))
        return last_step * factor

def ode_rk45(ode_fun: Callable, t_span: torch.Tensor, y0: torch.Tensor, 
             options: Optional[Dict[str, Any]] = None) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    """
    Modified rk45
    
    Parameters:
    -----------
    ode_fun : callable
        ODE function: f(t, y) 
    t_span : torch.Tensor
        Time span [t0, t1, ..., tfinal]
    y0 : torch.Tensor
        Initial conditions
    options : dict
        Options dictionary with keys:
        - rel_tol = 1e-3 ：relative tolerence
        - abs_tol = 1e-6 ：absolute tolerence
        - waitbar = True : whether to show progress
        - max_consecutive_failures = 10: Maximum number of consecutive step failures
    
    Returns:
    --------
    t : torch.Tensor
        Time points
    y : torch.Tensor
        Solution values
    stats : dict
        Statistics (nfevals)
    err_info : dict
        Error history and max step error
    """
    solver= 'RK45'
    sf=eq_solver(ode_fun, t_span, y0, solver, options)               
    bar=Wait_bar(t_span, sf.waitbar)  # Initialize the progress bar
    t, ang, stats, error_info =sf.run(bar)
    return t, ang, stats, error_info

def ode_sde_em(f: Callable, # function
                t_span: torch.Tensor, 
                y0: torch.Tensor, 
                options: Optional[Dict[str, Any]] = None
        ) -> Tuple[torch.Tensor, torch.Tensor, Dict, Dict]:
    """
    Modified Euler-Maruyama method for SDEs.
    
    Parameters:
    -----------
    f : callable
        function: f0, g0 = f(t, y) with f0 the drift term and g0 the diffusion term
    t_span : torch.Tensor
        Time span for integration
    y0 : torch.Tensor
        Initial condition
    options : dict, optional
        Integration options

    Returns:
    --------
    t : torch.Tensor
        Time points at which output is given
    y : torch.Tensor 
        Solution at tout
    stats : dict
        Integration statistics
    err_info : dict
        Error information
    """
    if options is None:
        options = {}

    # Initialize options
    waitbar = options.get('waitbar', True)
    
    # Extract odeset options
    rtol = options.get('rel_tol', 1e-3)
    atol = options.get('abs_tol', 1e-6)
    norm_control = options.get('NormControl', 'off') == 'on'
    max_consecutive_failures = options.get('max_consecutive_failures', 10)
    # Initialize waitbar
    if waitbar:
        # Estimate total progress based on time span
        t0 = t_span[0].item()
        t_final = t_span[-1].item()
        total_progress = t_final - t0
        pbar = tqdm(total=total_progress, desc='ODE Integration', 
                unit='time', ncols=100, bar_format='{l_bar}{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}]')
        last_update_time = time.time()
        update_interval = 0.1  # Update progress bar every 0.1 seconds
    
    # Initialize solution storage
    t0 = t_span[0]
    t_final = t_span[-1]
    t_dir = torch.sign(t_final - t0)
    
    # Ensure y0 is 1D tensor
    original_shape = y0.shape
    y0 = y0.reshape(-1,1)
    n_eq = y0.shape[0]
    
    # Data type
    dtype = y0.dtype
    device = y0.device
    
    # Step size constraints
    h_min = 16 * torch.finfo(dtype).eps
    h_min=torch.tensor(h_min,dtype=dtype,device=device)
    safe_h_max = 16.0 * torch.finfo(dtype).eps * torch.max(torch.abs(t0), torch.abs(t_final))
    default_h_max = torch.max(torch.abs(t_final - t0), safe_h_max)
    h_max = torch.min(torch.abs(t_final - t0), 
                    torch.tensor(options.get('MaxStep', default_h_max.item()), dtype=dtype, device=device))
    threshold = torch.tensor(atol, dtype=dtype, device=device)
    if norm_control:
        norm_y = torch.norm(y0)
    else:
        norm_y = torch.tensor(0.0, dtype=dtype, device=device)
    
    t = t0.clone()
    y = y0.clone()
    
    # Output configuration
    n_t_span = t_span.shape[0]
    refine = options.get('Refine', 4)
    
    if n_t_span > 2:
        output_pos = 1  # output only at tspan points
    elif refine <= 1:
        output_pos = 2  # computed points, no refinement
    else:
        output_pos = 3  # computed points, with refinement
        S = torch.linspace(1/refine, 1 - 1/refine, refine - 1, dtype=dtype, device=device)
    
    # Initialize output arrays
    if n_t_span > 2:
        t_out = torch.zeros(n_t_span, dtype=dtype, device=device)
        y_out = torch.zeros(n_eq, n_t_span, dtype=dtype, device=device)
    else:
        chunk = min(max(100, 50 * refine), refine + (2**13) // n_eq)
        t_out = torch.zeros(chunk, dtype=dtype, device=device)
        y_out = torch.zeros(n_eq, chunk, dtype=dtype, device=device)
    
    n_out = 0
    t_out[n_out] = t
    y_out[:, n_out] = y.view(-1)
    
    error_history = []
    n_calls = 0
    n_steps = 0

    # Pi value
    t2pi=torch.tensor(2*math.pi,dtype=dtype,device=device)
    # Initial step size
    h = torch.min(h_max, torch.max(h_min, 0.1 * torch.abs(t_final - t0)))
    h_abs = torch.abs(h)
    # Initial function evaluation
    f1, g1 = f(t, y)
    noise_dim=g1.shape[1]
    n_calls += 1

    finished = False
    next_idx = 1  # for tspan output
    
    # Main integration loop
    consecutive_failures = 0
    integration_failed= False

    while not finished:
        h_abs = torch.min(h_max, torch.max(h_min, h_abs))
        h = t_dir * h_abs
        if 1.1 * h_abs >= torch.abs(t_final - t):
            h = t_final - t
            h_abs = torch.abs(h)
            finished = True
        
        no_failed = True
        W1=torch.randn(noise_dim,1,dtype=dtype,device=device)
        W2=torch.randn(noise_dim,1,dtype=dtype,device=device)
        W=W1+W2
        while True:
            y2 = y+ f1 * h/2 + g1 @ W1 * torch.sqrt(h_abs/2)
            t2 = t + h / 2
            f2, g2 = f(t2, y2)
            
            y_new = y + f2 *h/2 + g2 @ W2 * torch.sqrt(h_abs/2)
            t_new = t + h

            y_full = y + f1 * h + g1 @ W * torch.sqrt(h_abs)
            
            n_calls += 1
            
            fE= y_new - y_full
            if norm_control:
                norm_y_new = torch.norm(y_new)
                scaling_Factor = torch.max(torch.max(norm_y, norm_y_new), threshold)
                err = h_abs *torch.norm(fE) / scaling_Factor
            else: 
                scaling_Factor = torch.max(torch.max(torch.abs(y), torch.abs(y_new)), threshold)
                err = fE / scaling_Factor
                err = h_abs * torch.max(err.abs())

            err=err.item()
            # Step acceptance
            if err > rtol:
                if torch.abs(h_abs - h_min) <  0.2 * h_min:
                    consecutive_failures += 1
                    if consecutive_failures >= max_consecutive_failures:
                        if waitbar:
                            pbar.close()
                        warnings.warn(
                            f"Step size reached minimum hmin = {h_min.item():.2e} at t={t.item():.2e}, but still cannot satisfy tolerance. "
                            f"Current error: {err:.2e}, Required tolerance: {rtol:.2e}. "
                            f"This may indicate a stiff ODE or overly strict tolerances. "
                            f"Consider using a stiff solver or relaxing tolerances.",
                            RuntimeWarning,
                            stacklevel=2
                        )
                        finished = True
                        integration_failed = True
                        break
                else:
                    consecutive_failures = 0  # Reset if we're still above hmin
                # Adaptive mode: shrink step and retry
                if no_failed:
                    no_failed = False
                    h_abs = torch.max(h_min, h_abs * max(0.1, 0.8 * (rtol / err) ** (1/1.5)))
                else:
                    h_abs = torch.max(h_min, 0.5 * h_abs)
                h = t_dir * h_abs
                finished = False
            else:
                # Accept step
                error_history.append(err)
                consecutive_failures = 0
                break
        n_steps += 1
        if integration_failed:
            break

        # Update waitbar if enabled
        if waitbar:
            current_time = time.time()
            if current_time - last_update_time >= update_interval or finished:
                progress = t_new.item() - t0.item()
                pbar.n = min(progress, total_progress)
                pbar.refresh()
                last_update_time = current_time
        
        # Output processing
        if output_pos == 2:  # computed points, no refinement
            n_out_new = 1
            t_out_new = t_new.unsqueeze(0)
            y_out_new = y_new.unsqueeze(1)
        elif output_pos == 3:  # computed points, with refinement
            t_ref = t + (t_new - t) * S
            n_out_new = refine
            t_out_new = torch.cat([t_ref, t_new.unsqueeze(0)])
            y_ntrp = ntrp_em(t_ref, t, y, h, y2, y_new)
            y_out_new = torch.cat([y_ntrp, y_new.unsqueeze(1)], dim=1) 
        else:  # output only at tspan points
            n_out_new = 0
            t_out_new = torch.tensor([], dtype=dtype, device=device)
            y_out_new = torch.tensor([], dtype=dtype, device=device)
            
            while next_idx < n_t_span:
                if t_dir * (t_new - t_span[next_idx]) < 0:
                    break
                n_out_new += 1
                t_out_new = torch.cat([t_out_new, t_span[next_idx].unsqueeze(0)])
                if t_span[next_idx] == t_new:
                    y_out_new = torch.cat([y_out_new, y_new], dim=1)
                else:
                    y_ntrp = ntrp_em(t_span[next_idx].unsqueeze(0), t, y, h, y2, y_new)
                    y_out_new = torch.cat([y_out_new, y_ntrp], dim=1)
                next_idx += 1
        y_out_new = y_out_new % t2pi
        # Store output
        if n_out_new > 0:
            old_n_out = n_out
            n_out = n_out + n_out_new
            
            # Expand arrays if needed
            if n_out+1 > t_out.shape[0]:
                extra = max(chunk, n_out_new)
                t_out_new_temp = torch.zeros(t_out.shape[0] + extra, dtype=dtype, device=device)
                t_out_new_temp[:t_out.shape[0]] = t_out
                t_out = t_out_new_temp
                
                y_out_new_temp = torch.zeros(n_eq, y_out.shape[1] + extra, dtype=dtype, device=device)
                y_out_new_temp[:, :y_out.shape[1]] = y_out
                y_out = y_out_new_temp
            
            t_out[old_n_out+1:n_out+1] = t_out_new
            y_out[:, old_n_out+1:n_out+1] = y_out_new
        
        # Step size adjustment for adaptive mode
        if  no_failed:
            temp = 1.25 * (err / rtol) ** (1/1.5)
            if temp > 0.2:
                h_abs = h_abs / temp
            else:
                h_abs = 5.0 * h_abs
        
        # Advance integration
        t = t_new
        y = y_new % t2pi
        if norm_control:
            norm_y = norm_y_new
        f1, g1 = f(t,y)
        n_calls += 1
    # Close waitbar
    if waitbar:
        pbar.n = total_progress
        pbar.refresh()
        pbar.close()

    # Prepare outputs
    t_out = t_out[:n_out+1]
    y_out = y_out[:, :n_out+1]
    
    stats = {'n_fevals': n_calls,
             'n_steps': n_steps,
             'n_output': n_out+1,
             'intergration': not integration_failed}
    err_info = {
        'err_history': error_history,
        'max_step_error': max(error_history) if error_history else 0.0
    }
    return t_out,y_out,stats,err_info

def ntrp_em(tinterp: torch.Tensor, t: torch.Tensor, y: torch.Tensor, 
            h: torch.Tensor, y_mid: torch.Tensor, y_end: torch.Tensor) -> torch.Tensor:
    """
    2nd order Interpolation for Euler-Maruyama method.
    """
    dtype = y.dtype
    device = y.device
    
    s = (tinterp - t) / h
    s = s.reshape(-1)

    l0 = (2.0*s-1.0)*(s-1.0)
    l_mid = -4.0*s*(s-1.0)
    l_end = s*(2.0*s-1.0)
    
    yinterp = l0*y + l_mid*y_mid + l_end*y_end
    
    return yinterp


import magbox
import numpy as np
import os
import pytest

testdata = [ # N, K, J, dt ,T
    (256,1.0,0.5,1,400),
    (512,1.0,0.5,1,400),
    (1024,1.0,0.5,1,400),
    (256,1.0,1.0,0.5,400),
    (256,1.0,2.0,0.1,400),
    (256,1.0,3.0,0.1,400),
]

def plot_fun(t,x,y,ft_abs,q,w,K,J,dispersion,dispersion_theory,err,mean_err,max_err):
    if "PYTEST_CURRENT_TEST" in os.environ:
        print("测试模式不作图")
        return
    import matplotlib.pyplot as plt
    plt.figure(figsize=(10,10))

    plt.subplot(2,2,1)
    plt.imshow(ft_abs.T,
            origin='lower',
            extent=(q[0],q[-1],w[0],w[-1]),
    )
    plt.ylim(0,1.5*(K+2*J))

    plt.subplot(2,2,2)
    plt.scatter(q,dispersion,label="Simulation")
    plt.plot(q,dispersion_theory,label="Theory")
    plt.legend()

    plt.subplot(2,2,3)
    plt.scatter(q,err,label="Error")
    plt.legend()
    plt.title(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")

    plt.subplot(2,2,4)
    plt.plot(t,x[0,:],label="x")
    plt.plot(t,y[0,:],label="y")
    plt.legend()
    # plt.title(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")
    plt.show()

@pytest.mark.parametrize("N,K,J,dt,T",testdata)
def test_spin_chain(N,K,J,dt,T):
    # N=256
    # K=1.0
    # J=0.5
    # dt=1
    # T=50
    rng=np.random.default_rng()
    theta0=np.ones(N)*0.01
    phi0=rng.random(N)*2*np.pi
    # phi0=10*np.arange(N)/N *2*np.pi
    LT = magbox.Lattice(type="square", size=[N], periodic=True)
    vars = magbox.Vars(K1=K, J=J)
    dispersion_fun=lambda qf: K+J*(1-np.cos(qf))*np.cos(np.mean(theta0))

    spin=magbox.spin(theta0,phi0,LT,device='cpu',type='f64')
    sf=magbox.llg(spin,vars,alpha=0,T=T,dt=dt)

    t_tc,ang,stats,err_info=sf.run(spin)
    t=t_tc.cpu().detach().numpy()
    theta=ang[::2].cpu().detach().numpy()
    phi=ang[1::2].cpu().detach().numpy()

    x=np.sin(theta)*np.cos(phi)
    y=np.sin(theta)*np.sin(phi)
    z=np.cos(theta)

    u=x+1j*y
    ft=np.fft.fft2(u)
    ft_abs=np.abs(ft)
    w=np.fft.fftfreq(len(t), dt)*2*np.pi
    q=np.fft.fftfreq(N,1)*2*np.pi

    ft_abs=np.fft.fftshift(ft_abs)
    w=np.fft.fftshift(w)
    q=np.fft.fftshift(q)

    dispersion=np.zeros(len(q))
    for idx in range(len(q)):
        arg_max=np.argmax(ft_abs[idx,:])
        dispersion[idx]=w[arg_max]
    dispersion_theory=dispersion_fun(q)
    err=dispersion/dispersion_theory-1
    max_err=np.max(np.abs(err))
    mean_err=np.mean(np.abs(err))

    print(f"mean error: {mean_err:.2e}, max err: {max_err:.2e}")

    plot_fun(t,x,y,ft_abs,q,w,K,J,dispersion,dispersion_theory,err,mean_err,max_err)

    assert mean_err<1e-2

if __name__=="__main__":
    test_spin_chain(256,3.0,3.5,0.1,300)


# spin_chain_test(256,1,0.5,1,50)



from .heff import heff
import torch 
from . import boxlib
from .initial import Lattice, Vars
from .spin import spin
import warnings

class llg:
    def __init__(self,sp:spin,vars:Vars=Vars(),gamma=1, alpha=0.01, Temp=0., dt=0.1, T=50):

        warnings.filterwarnings("ignore", message="Sparse CSR tensor support is in beta state")
        data_type=sp.data_type
        device=sp.device
        self.num=sp.num
        self.gamma=torch.tensor(gamma,dtype=data_type,device=device)
        self.alpha=torch.tensor(alpha,dtype=data_type,device=device)
        self.Temp=torch.tensor(Temp,dtype=data_type,device=device)
        self.dt=torch.tensor(dt,dtype=data_type,device=device)
        self.T=torch.tensor(T,dtype=data_type,device=device)
        self.tspan=torch.linspace(0,self.T,int(self.T/self.dt)+1,dtype=data_type,device=device)
        self.h_fun=heff(sp,vars)
        self.data_type=sp.data_type
        self.device=sp.device

        # strength for thermal field
        self.thermal_strength=torch.sqrt(2*self.alpha*self.Temp/self.gamma)

        # prepare sparse matrix index for BSR format
        self.block_crow=torch.arange(0,self.num+1,dtype=torch.int64,device=device)
        self.block_col=torch.arange(0,self.num,dtype=torch.int64,device=device)

        # prepare sparse matrix index for CSR format 
        '''
        format:[[*,*,*]
                [*,*,0]]
        '''
        odd_pos = torch.arange((self.num*2 + 2) // 2,dtype=torch.int64,device=device) * 5
        even_pos = odd_pos[:-1] + 3
        self.convert_csr_crow=torch.zeros(self.num*2+1, dtype=torch.int64,device=device)
        self.convert_csr_crow[0::2] = odd_pos
        self.convert_csr_crow[1::2] = even_pos

        total_length = 5 * self.num
        group_indices = torch.arange(total_length,dtype=torch.int64,device=device) // 5
        pos_in_group = torch.arange(total_length,dtype=torch.int64,device=device) % 5
        col_pattern=torch.tensor([0,1,2,0,1],dtype=torch.int64,device=device)
        self.convert_csr_col=group_indices * 3 + col_pattern[pos_in_group]

        # prepare sparse matrix index for CSR format for drift cart to sph
        '''
        format:[[*,*,*]
                [*,*,*]]
        '''
        self.drift_cart_to_sph_csr_crow=torch.arange(0,self.num*6+1,step=3,dtype=torch.int64,device=device)
        
        total_length = 6 * self.num
        group_indices = torch.arange(total_length,dtype=torch.int64,device=device) // 6
        pos_in_group = torch.arange(total_length,dtype=torch.int64,device=device) % 6
        col_pattern=torch.tensor([0,1,2,0,1,2],dtype=torch.int64,device=device)
        self.drift_cart_to_sph_csr_col=group_indices * 3 + col_pattern[pos_in_group]

    def llg_kernal(self,s_theta):
        cscv=1/s_theta
        cscv2=cscv**2
        value=self.gamma/(1+self.alpha**2)*torch.cat([-self.alpha*torch.ones(self.num,1,dtype=self.data_type,device=self.device),-cscv,cscv,-self.alpha*cscv2],1)
        return torch.sparse_bsr_tensor(self.block_crow,self.block_col,value.reshape(-1,2,2),(self.num*2,self.num*2))
    
    def llg_convert_bsr(self,sp):
        value=-torch.cat([sp.c_theta*sp.c_phi, sp.c_theta*sp.s_phi, -sp.s_theta,
                          -sp.s_theta*sp.s_phi , sp.s_theta*sp.c_phi,torch.zeros(sp.num,1,dtype=sp.data_type,device=sp.device)],1)
        return torch.sparse_bsr_tensor(self.block_crow,self.block_col,value.reshape(-1,2,3),(self.num*2,self.num*3),dtype=sp.data_type,device=sp.device)
    def llg_convert(self,s_theta,c_theta,s_phi,c_phi):
        value=-torch.cat([c_theta*c_phi, c_theta*s_phi, -s_theta,
                            -s_theta*s_phi , s_theta*c_phi],1)
        return torch.sparse_csr_tensor(self.convert_csr_crow,self.convert_csr_col,value.reshape(-1),(self.num*2,self.num*3),dtype=self.data_type,device=self.device)
    
    def llg_drift_cart_to_sph(self,s_theta,c_theta,s_phi,c_phi):
        cot_theta=c_theta/s_theta
        csc_theta=1/s_theta
        value=self.gamma/(1+self.alpha**2)*torch.cat(
            [self.alpha*c_theta*c_phi-s_phi, c_phi+self.alpha*c_theta*s_phi, -self.alpha*s_theta,
             -self.alpha*csc_theta*s_phi-cot_theta*c_phi, -cot_theta*s_phi + self.alpha*csc_theta*c_phi, torch.ones(self.num,1,dtype=self.data_type,device=self.device)],1)
        return torch.sparse_csr_tensor(self.drift_cart_to_sph_csr_crow,self.drift_cart_to_sph_csr_col,value.reshape(-1), (self.num*2, self.num*3),dtype=self.data_type,device=self.device)

    def llg_drift(self,t, ang):
        theta=ang[0::2]
        phi=ang[1::2]

        s_theta=torch.sin(theta)
        c_theta=torch.cos(theta)
        s_phi=torch.sin(phi)
        c_phi=torch.cos(phi)

        # kernal=self.llg_kernal(s_theta)
        # h3_to_h2=self.llg_convert(s_theta,c_theta,s_phi,c_phi)
        kernal=self.llg_drift_cart_to_sph(s_theta,c_theta,s_phi,c_phi)
        h3=self.h_fun.all3(s_theta,c_theta,s_phi,c_phi)
        drift_core=kernal @ h3
        return drift_core
    def llg_thermal(self, t, ang):
        theta=ang[0::2]
        phi=ang[1::2]

        s_theta=torch.sin(theta)
        c_theta=torch.cos(theta)
        s_phi=torch.sin(phi)
        c_phi=torch.cos(phi)

        # kernal=self.llg_kernal(s_theta)
        # h3_to_h2=self.llg_convert(s_theta,c_theta,s_phi,c_phi)
        kernal=self.llg_drift_cart_to_sph(s_theta,c_theta,s_phi,c_phi)
        h3=self.h_fun.all3(s_theta,c_theta,s_phi,c_phi)
        correction=self.Stratonovich_correction(s_theta,c_theta)
        drift_core=kernal @ h3
        return drift_core-correction, kernal*self.thermal_strength

    def Stratonovich_correction(self,s_theta,c_theta):
        cot_theta=c_theta/s_theta
        value=self.gamma * self.alpha * self.Temp/(1+self.alpha**2)*torch.cat([cot_theta, torch.zeros(self.num,1,dtype=self.data_type,device=self.device)],1)
        return value.reshape(-1,1)
    def run(self,sp):
        
        
        ini=sp.ang
        odeset={"rel_tol":max(self.alpha.item()*1e-2,1e-4),"abs_tol":max(self.alpha.item()*1e-4,1e-6)}
        if self.Temp==0:
            llg_fun=self.llg_drift
            t,ang,stats,erro_info=boxlib.ode_rk45(llg_fun, self.tspan, ini, options=odeset)
        else:
            llg_fun=self.llg_thermal
            t,ang,stats,erro_info=boxlib.ode_sde_em(llg_fun, self.tspan,ini, options=odeset)
        return t,ang,stats,erro_info



from magbox import llg,spin, Lattice
from torch import sin, cos
from matplotlib import pyplot as plt
import numpy as np
import cProfile, pstats, io

pr=cProfile.Profile()
rng=np.random.default_rng()

N1=32
N2=32
lt=Lattice(type="square",size=[N1,N2],periodic=True)
theta0=np.acos(rng.random([N1,N2])*2-1)
phi0=rng.random([N1,N2])*2*np.pi
# theta0=np.ones([N1])*0.1
# phi0=np.zeros([N1])
alpha=0.1

pr.enable()
sp=spin(theta0,phi0,lattice_type=lt,device="gpu",type="f32")
sf=llg(sp,dt=1,alpha=alpha,T=200)
ang=sp.ang

t,ang,stats,erro_info=sf.run(sp)
theta=ang[0::2]
phi=ang[1::2]
x=sin(theta)*cos(phi)
y=sin(theta)*sin(phi)
z=cos(theta)

t_np=t.cpu().detach().numpy()
x_np=x.cpu().detach().numpy()
y_np=y.cpu().detach().numpy()
z_np=z.cpu().detach().numpy()
z0=z_np[0,0]
zt=1+np.exp(-2*alpha*t_np)*(z0-1)
pr.disable()

s = io.StringIO()
sortby = "cumtime"  
ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
ps.print_stats()
# print(s.getvalue())
pr.dump_stats("llg_gpu_f32_no_lambda.prof")

plt.figure(figsize=(6,6))

plt.subplot(2,1,1)
plt.plot(t_np,x_np[0,:],label="m_x")
plt.plot(t_np,y_np[0,:],label="m_y")
plt.xlabel("Time")
plt.title("m_x and m_y")
plt.legend()

plt.subplot(2,1,2)
plt.plot(t_np,z_np[0,:],label="m_z")
# plt.plot(t_np,zt,label="m_z")
plt.xlabel("Time")
plt.title("m_z")
plt.legend()

plt.tight_layout()
plt.show()





from tqdm import tqdm
import time

class Wait_bar:
    def __init__(self, t_span, flag):
        self.t0=0
        if flag:
            self.t0 = t_span[0].item()
            self.t_final = t_span[-1].item()
            self.total_progress = self.t_final - self.t0
            self.pbar = tqdm(total=self.total_progress, desc='ODE Integration', 
                    unit='time', ncols=100, bar_format='{l_bar}{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}]')
            self.last_update_time = time.time()
            self.update_interval = 0.1  # Update progress bar every 0.1 seconds
    def update(self, t, flag, done):
        if flag:
            current_time = time.time()
            if current_time - self.last_update_time >= self.update_interval or done:
                progress = t.item() - self.t0
                self.pbar.n = min(progress,self.total_progress)
                self.pbar.refresh()
                self.last_update_time = current_time
    def close(self,flag):
        if flag: 
            self.pbar.close()


import torch
from . import boxlib
from .initial import Vars, Lattice
from .spin import spin
from dataclasses import replace

class heff:
    def __init__(self, sp:spin, vars:Vars): # B=0,B_dir=[0,0,1],J=1,K_dir=[0,0,1],K1=1,D=0
        # self.spin = spin

        B=vars.B # external field
        B_dir=vars.B_dir # external field direction

        J=vars.J # exchange interaction
        K1_dir=vars.K1_dir # uniaxial anisotropy axis
        K1=vars.K1 # uniaxial anisotropy strength
        D=vars.D # Dzyaloshinskii-Moriya interaction

        self.num=sp.num
        self.B=torch.tensor(B,dtype=sp.data_type,device=sp.device)# external field
        self.B_dir=torch.tensor(B_dir,dtype=sp.data_type,device=sp.device)
        self.B_dir=self.B_dir.view(3)
        self.B_dir=self.B_dir/torch.norm(self.B_dir)
        self.B=self.B*self.B_dir

        self.K_dir=torch.tensor(K1_dir,dtype=sp.data_type,device=sp.device)
        self.K_dir=self.K_dir.view(3)
        self.K_dir=self.K_dir/torch.norm(self.K_dir)
        self.K1=torch.tensor(K1,dtype=sp.data_type,device=sp.device)
        self.K1=self.K1*self.K_dir

        self.H = torch.zeros((self.num,3))
        self.data_type=sp.data_type
        self.device=sp.device

        l_type=sp.lattice_type
        N_dim=l_type.size
        N_dim=len(N_dim)

        if isinstance(J,(int,float)) or (isinstance(J,list) and len(J)==1):
            J_val=J
            self.Jmtx=J_val*boxlib.get_Jmtx(replace(l_type,J_direction=None), device=sp.device,data_type=sp.data_type)
        elif isinstance(J,list):
            if len(J)!=N_dim:
                raise ValueError(f"J should be a scalar or a list of length {N_dim}")
            # self.Jmtx=J.to(device=spin.device,dtype=spin.data_type)
            self.Jmtx=torch.zeros((self.num,self.num),dtype=sp.data_type,device=sp.device)
            for i in range(N_dim):
                self.Jmtx+=J[i]*boxlib.get_Jmtx(replace(l_type,J_direction=i), device=sp.device,data_type=sp.data_type)
    def energy(self,ang):
        theta=ang[0::2]
        phi=ang[1::2]
        s_theta=torch.sin(theta)
        c_theta=torch.cos(theta)
        s_phi=torch.sin(phi)
        c_phi=torch.cos(phi)
        t_num=theta.shape[1]

        cartS=self.get_cart_S(s_theta,c_theta,s_phi,c_phi)

        E_zeeman= - torch.sum(self.B[0] * cartS[:,0:t_num]+self.B[1] * cartS[:,t_num:2*t_num]+self.B[2] * cartS[:,2*t_num:], dim=0)
        E_exch= -0.5 * torch.sum(cartS * (self.Jmtx @ cartS), dim=0)
        E_exch=E_exch[0::3] + E_exch[1::3] + E_exch[2::3]
        E_anis= -0.5 * torch.norm(self.K1) * torch.sum((cartS[:,0:t_num] *self.K_dir[0]+cartS[:,t_num:2*t_num] *self.K_dir[1]+cartS[:,2*t_num:] *self.K_dir[2])**2, dim=0)
        E_total=E_zeeman + E_exch + E_anis
        return E_total
    def zeeman3(self):
        return self.B*torch.ones((self.num,3),dtype=self.data_type,device=self.device)
    # def zeeman2(self,ctheta,stheta,cphi,sphi):
    #     h_theta=self.B_dir@torch.cat([])
    # return self.B*torch.cat([*],1)

    def dipole3(self):
        # calculate dipole field
        return torch.zeros((self.num,3))

    def exchange3(self,cartS):
        # calculate exchange field
        return self.Jmtx @ cartS
    def uni_anisotropy3(self,cartS):
        # calculate anisotropy field
        return self.K1 * cartS
    def DM3(self):
        # calculate DM field
        return torch.zeros((self.num,3))
    def all3(self, s_theta,c_theta,s_phi,c_phi):
        cartS=self.get_cart_S(s_theta,c_theta,s_phi,c_phi)
        h0=self.zeeman3() + self.exchange3(cartS) + self.uni_anisotropy3(cartS)
        return h0.reshape(-1,1)
    
    @staticmethod
    def get_cart_S(s_theta,c_theta,s_phi,c_phi):
        x = s_theta * c_phi
        y = s_theta * s_phi
        z = c_theta
        return torch.cat([x, y, z], dim=1)
    
        


import importlib.metadata
__version__=importlib.metadata.version(__name__)
__author__="Yutian Wang"

from .llg import llg
from .heff import heff
from .spin import spin
from .boxlib import get_data_type
from .initial import Lattice, Vars